@@ -36,7 +36,7 @@ ldns_lookup_table ldns_error_str[] = {
         { LDNS_STATUS_NETWORK_ERR, "Could not send or receive, because of network error" },
         { LDNS_STATUS_ADDRESS_ERR, "Could not start AXFR, because of address error" },
         { LDNS_STATUS_FILE_ERR, "Could not open the files" },
-        { LDNS_STATUS_UNKNOWN_INET, "Uknown address family" },
+        { LDNS_STATUS_UNKNOWN_INET, "Unknown address family" },
         { LDNS_STATUS_NOT_IMPL, "This function is not implemented (yet), please notify the developers - or not..." },
 	{ LDNS_STATUS_NULL, "Supplied value pointer null" },
         { LDNS_STATUS_CRYPTO_UNKNOWN_ALGO, "Unknown cryptographic algorithm" },
