@@ -12,7 +12,7 @@
 
 int verbosity = 0;
 
-int
+static int
 usage(FILE *fp, char *prog) {
 	fprintf(fp, "%s [options] domain\n", prog);
 	fprintf(fp, "  print out the owner names for domain and the record types for those names\n");
@@ -27,7 +27,7 @@ usage(FILE *fp, char *prog) {
 	return 0;
 }
 
-ldns_rdf *
+static ldns_rdf *
 create_dname_plus_1(ldns_rdf *dname)
 {
 	uint8_t *wire;
@@ -94,7 +94,7 @@ create_dname_plus_1(ldns_rdf *dname)
 	return newdname;
 }
 
-ldns_rdf *
+static ldns_rdf *
 create_plus_1_dname(ldns_rdf *dname)
 {
 	ldns_rdf *label;
@@ -120,7 +120,7 @@ create_plus_1_dname(ldns_rdf *dname)
 	return label;
 }
 
-ldns_status
+static void 
 query_type_bitmaps(ldns_resolver *res, 
                    uint16_t res_flags,
                    const ldns_rdf *name,
@@ -191,11 +191,8 @@ query_type_bitmaps(ldns_resolver *res,
 				}
 			}
 		}
-		
 		pos += (uint16_t) bitmap_length;
 	}
-	
-	return LDNS_STATUS_OK;
 }
 
 
@@ -259,7 +256,7 @@ main(int argc, char *argv[])
 				full = true;
 			} else if (strncmp(argv[i], "-s", 3) == 0) {
 				if (i + 1 < argc) {
-					if (!ldns_str2rdf_dname(&startpoint, argv[i + 1]) == LDNS_STATUS_OK) {
+					if (ldns_str2rdf_dname(&startpoint, argv[i + 1]) != LDNS_STATUS_OK) {
 						printf("Bad start point name: %s\n", argv[i + 1]);
 						exit(1);
 					}
@@ -639,7 +636,7 @@ main(int argc, char *argv[])
 			/* ok, so now we know all the types present at this name,
 			 * query for those one by one (...)
 			 */
-			status = query_type_bitmaps(res, LDNS_RD, ldns_rr_owner(nsec_rr),
+			query_type_bitmaps(res, LDNS_RD, ldns_rr_owner(nsec_rr),
 			                   ldns_rr_rdf(nsec_rr, 1));
 			/* print this nsec and its signatures too */
 			ldns_rr_print(stdout, nsec_rr);
