@@ -28,6 +28,14 @@
 #include <time.h>
 #include <sys/time.h>
 
+#ifdef HAVE_SSL
+#include <openssl/bn.h>
+#include <openssl/rsa.h>
+#ifdef USE_DSA
+#include <openssl/dsa.h>
+#endif
+#endif
+
 #ifndef INET_ADDRSTRLEN
 #define INET_ADDRSTRLEN 16
 #endif
@@ -56,6 +64,12 @@ ldns_lookup_table ldns_algorithms[] = {
 #ifdef USE_ECDSA
         { LDNS_ECDSAP256SHA256, "ECDSAP256SHA256"},
         { LDNS_ECDSAP384SHA384, "ECDSAP384SHA384"},
+#endif
+#ifdef USE_ED25519
+	{ LDNS_ED25519, "ED25519"},
+#endif
+#ifdef USE_ED448
+	{ LDNS_ED448, "ED448"},
 #endif
         { LDNS_INDIRECT, "INDIRECT" },
         { LDNS_PRIVATEDNS, "PRIVATEDNS" },
@@ -469,9 +483,18 @@ ldns_rdf2buffer_str_str(ldns_buffer *output, const ldns_rdf *rdf)
 ldns_status
 ldns_rdf2buffer_str_b64(ldns_buffer *output, const ldns_rdf *rdf)
 {
-	size_t size = ldns_b64_ntop_calculate_size(ldns_rdf_size(rdf));
-	char *b64 = LDNS_XMALLOC(char, size);
-	if(!b64) return LDNS_STATUS_MEM_ERR;
+	size_t size;
+	char *b64;
+
+	if (ldns_rdf_size(rdf) == 0) {
+		ldns_buffer_printf(output, "0");
+		return ldns_buffer_status(output);
+	} else
+		size = ldns_b64_ntop_calculate_size(ldns_rdf_size(rdf));
+
+	if (!(b64 = LDNS_XMALLOC(char, size)))
+		return LDNS_STATUS_MEM_ERR;
+
 	if (ldns_b64_ntop(ldns_rdf_data(rdf), ldns_rdf_size(rdf), b64, size)) {
 		ldns_buffer_printf(output, "%s", b64);
 	}
@@ -511,7 +534,7 @@ ldns_rdf2buffer_str_hex(ldns_buffer *output, const ldns_rdf *rdf)
 	return ldns_buffer_status(output);
 }
 
-ldns_status
+static ldns_status
 ldns_rdf2buffer_str_type_fmt(ldns_buffer *output,
 		const ldns_output_format* fmt, const ldns_rdf *rdf)
 {
@@ -567,11 +590,7 @@ ldns_rdf2buffer_str_cert_alg(ldns_buffer *output, const ldns_rdf *rdf)
 ldns_status
 ldns_rdf2buffer_str_alg(ldns_buffer *output, const ldns_rdf *rdf)
 {
-	/* don't use algorithm mnemonics in the presentation format
-	   this kind of got sneaked into the rfc's */
-        uint8_t data = ldns_rdf_data(rdf)[0];
-		ldns_buffer_printf(output, "%d", data);
-	return ldns_buffer_status(output);
+	return ldns_rdf2buffer_str_int8(output, rdf);
 }
 
 static void
@@ -598,29 +617,28 @@ ldns_rr_type2buffer_str(ldns_buffer *output, const ldns_rr_type type)
 
 	descriptor = ldns_rr_descript(type);
 
-	if (descriptor && descriptor->_name) {
-		ldns_buffer_printf(output, "%s", descriptor->_name);
-	} else {
-		/* exceptions for pseudotypes */
-		switch (type) {
-			case LDNS_RR_TYPE_IXFR:
-				ldns_buffer_printf(output, "IXFR");
-				break;
-			case LDNS_RR_TYPE_AXFR:
-				ldns_buffer_printf(output, "AXFR");
-				break;
-			case LDNS_RR_TYPE_MAILA:
-				ldns_buffer_printf(output, "MAILA");
-				break;
-			case LDNS_RR_TYPE_MAILB:
-				ldns_buffer_printf(output, "MAILB");
-				break;
-			case LDNS_RR_TYPE_ANY:
-				ldns_buffer_printf(output, "ANY");
-				break;
-			default:
+	switch (type) {
+		case LDNS_RR_TYPE_IXFR:
+			ldns_buffer_printf(output, "IXFR");
+			break;
+		case LDNS_RR_TYPE_AXFR:
+			ldns_buffer_printf(output, "AXFR");
+			break;
+		case LDNS_RR_TYPE_MAILA:
+			ldns_buffer_printf(output, "MAILA");
+			break;
+		case LDNS_RR_TYPE_MAILB:
+			ldns_buffer_printf(output, "MAILB");
+			break;
+		case LDNS_RR_TYPE_ANY:
+			ldns_buffer_printf(output, "ANY");
+			break;
+		default:
+			if (descriptor && descriptor->_name) {
+				ldns_buffer_printf(output, "%s", descriptor->_name);
+			} else {
 				ldns_buffer_printf(output, "TYPE%u", type);
-		}
+			}
 	}
 	return ldns_buffer_status(output);
 }
@@ -820,7 +838,7 @@ ldns_rdf2buffer_str_wks(ldns_buffer *output, const ldns_rdf *rdf)
 #endif
 
 	for (current_service = 0;
-	     current_service < ldns_rdf_size(rdf) * 7; current_service++) {
+	     current_service < (ldns_rdf_size(rdf)-1)*8; current_service++) {
 		if (ldns_get_bit(&(ldns_rdf_data(rdf)[1]), current_service)) {
 			service = getservbyport((int) htons(current_service),
 			                        proto_name);
@@ -837,7 +855,7 @@ ldns_rdf2buffer_str_wks(ldns_buffer *output, const ldns_rdf *rdf)
 	return ldns_buffer_status(output);
 }
 
-ldns_status
+static ldns_status
 ldns_rdf2buffer_str_nsec_fmt(ldns_buffer *output,
 		const ldns_output_format* fmt, const ldns_rdf *rdf)
 {
@@ -1084,12 +1102,12 @@ ldns_rdf2buffer_str_ipseckey(ldns_buffer *output, const ldns_rdf *rdf)
 			/* no gateway */
 			break;
 		case 1:
-			gateway_data = LDNS_XMALLOC(uint8_t, LDNS_IP4ADDRLEN);
-                        if(!gateway_data)
-                                return LDNS_STATUS_MEM_ERR;
 			if (ldns_rdf_size(rdf) < offset + LDNS_IP4ADDRLEN) {
 				return LDNS_STATUS_ERR;
 			}
+			gateway_data = LDNS_XMALLOC(uint8_t, LDNS_IP4ADDRLEN);
+                        if(!gateway_data)
+                                return LDNS_STATUS_MEM_ERR;
 			memcpy(gateway_data, &data[offset], LDNS_IP4ADDRLEN);
 			gateway = ldns_rdf_new(LDNS_RDF_TYPE_A,
 					LDNS_IP4ADDRLEN , gateway_data);
@@ -1100,12 +1118,12 @@ ldns_rdf2buffer_str_ipseckey(ldns_buffer *output, const ldns_rdf *rdf)
                         }
 			break;
 		case 2:
-			gateway_data = LDNS_XMALLOC(uint8_t, LDNS_IP6ADDRLEN);
-                        if(!gateway_data)
-                                return LDNS_STATUS_MEM_ERR;
 			if (ldns_rdf_size(rdf) < offset + LDNS_IP6ADDRLEN) {
 				return LDNS_STATUS_ERR;
 			}
+			gateway_data = LDNS_XMALLOC(uint8_t, LDNS_IP6ADDRLEN);
+                        if(!gateway_data)
+                                return LDNS_STATUS_MEM_ERR;
 			memcpy(gateway_data, &data[offset], LDNS_IP6ADDRLEN);
 			offset += LDNS_IP6ADDRLEN;
 			gateway =
@@ -1128,12 +1146,13 @@ ldns_rdf2buffer_str_ipseckey(ldns_buffer *output, const ldns_rdf *rdf)
 	}
 
 	if (ldns_rdf_size(rdf) <= offset) {
+                ldns_rdf_deep_free(gateway);
 		return LDNS_STATUS_ERR;
 	}
 	public_key_size = ldns_rdf_size(rdf) - offset;
 	public_key_data = LDNS_XMALLOC(uint8_t, public_key_size);
         if(!public_key_data) {
-                ldns_rdf_free(gateway);
+                ldns_rdf_deep_free(gateway);
                 return LDNS_STATUS_MEM_ERR;
         }
 	memcpy(public_key_data, &data[offset], public_key_size);
@@ -1141,7 +1160,7 @@ ldns_rdf2buffer_str_ipseckey(ldns_buffer *output, const ldns_rdf *rdf)
 			public_key_size, public_key_data);
         if(!public_key) {
                 LDNS_FREE(public_key_data);
-                ldns_rdf_free(gateway);
+                ldns_rdf_deep_free(gateway);
                 return LDNS_STATUS_MEM_ERR;
         }
 
@@ -1153,8 +1172,8 @@ ldns_rdf2buffer_str_ipseckey(ldns_buffer *output, const ldns_rdf *rdf)
 	ldns_buffer_printf(output, " ");
 	(void) ldns_rdf2buffer_str(output, public_key);
 
-	ldns_rdf_free(gateway);
-	ldns_rdf_free(public_key);
+	ldns_rdf_deep_free(gateway);
+	ldns_rdf_deep_free(public_key);
 
 	return ldns_buffer_status(output);
 }
@@ -1217,7 +1236,7 @@ ldns_rdf2buffer_str_tag(ldns_buffer *output, const ldns_rdf *rdf)
 	chars = ldns_rdf_data(rdf) + 1;
 	while (nchars > 0) {
 		ch = (char)*chars++;
-		if (! isalnum(ch)) {
+		if (! isalnum((unsigned char)ch)) {
 			return LDNS_STATUS_WIRE_RDATA_ERR;
 		}
 		ldns_buffer_printf(output, "%c", ch);
@@ -1280,7 +1299,95 @@ ldns_rdf2buffer_str_hip(ldns_buffer *output, const ldns_rdf *rdf)
 	return ldns_buffer_status(output);
 }
 
+/* implementation mimiced from ldns_rdf2buffer_str_ipseckey */
 ldns_status
+ldns_rdf2buffer_str_amtrelay(ldns_buffer *output, const ldns_rdf *rdf)
+{
+	/* wire format from
+	 * draft-ietf-mboned-driad-amt-discovery Section 4.2  
+	 */
+	uint8_t *data = ldns_rdf_data(rdf);
+	uint8_t precedence;
+	uint8_t discovery_optional;
+	uint8_t relay_type;
+
+	ldns_rdf *relay = NULL;
+	uint8_t *relay_data;
+
+	size_t offset = 0;
+	ldns_status status;
+
+	if (ldns_rdf_size(rdf) < 2) {
+		return LDNS_STATUS_WIRE_RDATA_ERR;
+	}
+	precedence = data[0];
+	discovery_optional = ((data[1] & 0x80) >> 7);
+	relay_type = data[1] & 0x7F;
+	offset = 2;
+
+	switch (relay_type) {
+		case 0:
+			/* no relay */
+			break;
+		case 1:
+			if (ldns_rdf_size(rdf) < offset + LDNS_IP4ADDRLEN) {
+				return LDNS_STATUS_ERR;
+			}
+			relay_data = LDNS_XMALLOC(uint8_t, LDNS_IP4ADDRLEN);
+                        if(!relay_data)
+                                return LDNS_STATUS_MEM_ERR;
+			memcpy(relay_data, &data[offset], LDNS_IP4ADDRLEN);
+			relay = ldns_rdf_new(LDNS_RDF_TYPE_A,
+					LDNS_IP4ADDRLEN , relay_data);
+			offset += LDNS_IP4ADDRLEN;
+                        if(!relay) {
+                                LDNS_FREE(relay_data);
+                                return LDNS_STATUS_MEM_ERR;
+                        }
+			break;
+		case 2:
+			if (ldns_rdf_size(rdf) < offset + LDNS_IP6ADDRLEN) {
+				return LDNS_STATUS_ERR;
+			}
+			relay_data = LDNS_XMALLOC(uint8_t, LDNS_IP6ADDRLEN);
+                        if(!relay_data)
+                                return LDNS_STATUS_MEM_ERR;
+			memcpy(relay_data, &data[offset], LDNS_IP6ADDRLEN);
+			offset += LDNS_IP6ADDRLEN;
+			relay =
+				ldns_rdf_new(LDNS_RDF_TYPE_AAAA,
+						LDNS_IP6ADDRLEN, relay_data);
+                        if(!relay) {
+                                LDNS_FREE(relay_data);
+                                return LDNS_STATUS_MEM_ERR;
+                        }
+			break;
+		case 3:
+			status = ldns_wire2dname(&relay, data,
+					ldns_rdf_size(rdf), &offset);
+                        if(status != LDNS_STATUS_OK)
+                                return status;
+			break;
+		default:
+			/* error? */
+			break;
+	}
+
+	if (ldns_rdf_size(rdf) != offset) {
+                ldns_rdf_deep_free(relay);
+		return LDNS_STATUS_ERR;
+	}
+	ldns_buffer_printf(output, "%u %u %u ",
+			precedence, discovery_optional, relay_type);
+	if (relay)
+	  	(void) ldns_rdf2buffer_str(output, relay);
+
+	ldns_rdf_deep_free(relay);
+	return ldns_buffer_status(output);
+}
+
+
+static ldns_status
 ldns_rdf2buffer_str_fmt(ldns_buffer *buffer,
 		const ldns_output_format* fmt, const ldns_rdf *rdf)
 {
@@ -1294,7 +1401,11 @@ ldns_rdf2buffer_str_fmt(ldns_buffer *buffer,
 		case LDNS_RDF_TYPE_DNAME:
 			res = ldns_rdf2buffer_str_dname(buffer, rdf);
 			break;
-		case LDNS_RDF_TYPE_INT8:
+		case LDNS_RDF_TYPE_INT8: /* Don't output mnemonics for these */
+		case LDNS_RDF_TYPE_ALG:
+		case LDNS_RDF_TYPE_CERTIFICATE_USAGE:
+		case LDNS_RDF_TYPE_SELECTOR:
+		case LDNS_RDF_TYPE_MATCHING_TYPE:
 			res = ldns_rdf2buffer_str_int8(buffer, rdf);
 			break;
 		case LDNS_RDF_TYPE_INT16:
@@ -1345,9 +1456,6 @@ ldns_rdf2buffer_str_fmt(ldns_buffer *buffer,
 		case LDNS_RDF_TYPE_CERT_ALG:
 			res = ldns_rdf2buffer_str_cert_alg(buffer, rdf);
 			break;
-		case LDNS_RDF_TYPE_ALG:
-			res = ldns_rdf2buffer_str_alg(buffer, rdf);
-			break;
 		case LDNS_RDF_TYPE_UNKNOWN:
 			res = ldns_rdf2buffer_str_unknown(buffer, rdf);
 			break;
@@ -1394,6 +1502,9 @@ ldns_rdf2buffer_str_fmt(ldns_buffer *buffer,
 		case LDNS_RDF_TYPE_LONG_STR:
 			res = ldns_rdf2buffer_str_long_str(buffer, rdf);
 			break;
+		case LDNS_RDF_TYPE_AMTRELAY:
+			res = ldns_rdf2buffer_str_amtrelay(buffer, rdf);
+			break;
 		}
 	} else {
 		/** This will write mangled RRs */
@@ -1927,19 +2038,100 @@ ldns_gost_key2buffer_str(ldns_buffer *output, EVP_PKEY *p)
 }
 #endif
 
+#if defined(HAVE_SSL) && defined(USE_ED25519)
+static ldns_status
+ldns_ed25519_key2buffer_str(ldns_buffer *output, EVP_PKEY *p)
+{
+	unsigned char* pp = NULL;
+	int ret;
+	ldns_rdf *b64_bignum;
+	ldns_status status;
+
+	ldns_buffer_printf(output, "PrivateKey: ");
+
+	ret = i2d_PrivateKey(p, &pp);
+	/* 16 byte asn (302e020100300506032b657004220420) + 32byte key */
+	if(ret != 16 + 32) {
+		OPENSSL_free(pp);
+		return LDNS_STATUS_ERR;
+	}
+	b64_bignum = ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64,
+		(size_t)ret-16, pp+16);
+	status = ldns_rdf2buffer_str(output, b64_bignum);
+
+	ldns_rdf_deep_free(b64_bignum);
+	OPENSSL_free(pp);
+	ldns_buffer_printf(output, "\n");
+	return status;
+}
+#endif
+
+#if defined(HAVE_SSL) && defined(USE_ED448)
+static ldns_status
+ldns_ed448_key2buffer_str(ldns_buffer *output, EVP_PKEY *p)
+{
+	unsigned char* pp = NULL;
+	int ret;
+	ldns_rdf *b64_bignum;
+	ldns_status status;
+
+	ldns_buffer_printf(output, "PrivateKey: ");
+
+	ret = i2d_PrivateKey(p, &pp);
+	/* some-ASN + 57byte key */
+	if(ret != 16 + 57) {
+		OPENSSL_free(pp);
+		return LDNS_STATUS_ERR;
+	}
+	b64_bignum = ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64,
+		(size_t)ret-16, pp+16);
+	status = ldns_rdf2buffer_str(output, b64_bignum);
+
+	ldns_rdf_deep_free(b64_bignum);
+	OPENSSL_free(pp);
+	ldns_buffer_printf(output, "\n");
+	return status;
+}
+#endif
+
+#if defined(HAVE_SSL)
+/** print one b64 encoded bignum to a line in the keybuffer */
+static int
+ldns_print_bignum_b64_line(ldns_buffer* output, const char* label, const BIGNUM* num)
+{
+	unsigned char  *bignumbuf = LDNS_XMALLOC(unsigned char, LDNS_MAX_KEYLEN);
+	if(!bignumbuf) return 0;
+
+	ldns_buffer_printf(output, "%s: ", label);
+	if(num) {
+		ldns_rdf *b64_bignum = NULL;
+		int i = BN_bn2bin(num, bignumbuf);
+		if (i > LDNS_MAX_KEYLEN) {
+			LDNS_FREE(bignumbuf);
+			return 0;
+		}
+		b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, (size_t)i, bignumbuf);
+		if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
+			ldns_rdf_deep_free(b64_bignum);
+			LDNS_FREE(bignumbuf);
+			return 0;
+		}
+		ldns_rdf_deep_free(b64_bignum);
+		ldns_buffer_printf(output, "\n");
+	} else {
+		ldns_buffer_printf(output, "(Not available)\n");
+	}
+	LDNS_FREE(bignumbuf);
+	return 1;
+}
+#endif
+
 ldns_status
 ldns_key2buffer_str(ldns_buffer *output, const ldns_key *k)
 {
 	ldns_status status = LDNS_STATUS_OK;
 	unsigned char  *bignum;
 #ifdef HAVE_SSL
-#  ifndef S_SPLINT_S
-	uint16_t i;
-#  endif
-	/* not used when ssl is not defined */
-	/*@unused@*/
-	ldns_rdf *b64_bignum = NULL;
-
 	RSA *rsa;
 	DSA *dsa;
 #endif /* HAVE_SSL */
@@ -2009,132 +2201,43 @@ ldns_key2buffer_str(ldns_buffer *output, const ldns_key *k)
 
 				/* print to buf, convert to bin, convert to b64,
 				 * print to buf */
-				ldns_buffer_printf(output, "Modulus: ");
+
 #ifndef S_SPLINT_S
-				i = (uint16_t)BN_bn2bin(rsa->n, bignum);
-				if (i > LDNS_MAX_KEYLEN) {
-					goto error;
-				}
-				b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-				if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-					ldns_rdf_deep_free(b64_bignum);
-					goto error;
-				}
-				ldns_rdf_deep_free(b64_bignum);
-				ldns_buffer_printf(output, "\n");
-				ldns_buffer_printf(output, "PublicExponent: ");
-				i = (uint16_t)BN_bn2bin(rsa->e, bignum);
-				if (i > LDNS_MAX_KEYLEN) {
-					goto error;
-				}
-				b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-				if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-					ldns_rdf_deep_free(b64_bignum);
-					goto error;
-				}
-				ldns_rdf_deep_free(b64_bignum);
-				ldns_buffer_printf(output, "\n");
-
-				ldns_buffer_printf(output, "PrivateExponent: ");
-				if (rsa->d) {
-					i = (uint16_t)BN_bn2bin(rsa->d, bignum);
-					if (i > LDNS_MAX_KEYLEN) {
+				if(1) {
+					const BIGNUM *n=NULL, *e=NULL, *d=NULL,
+						*p=NULL, *q=NULL, *dmp1=NULL,
+						*dmq1=NULL, *iqmp=NULL;
+#if OPENSSL_VERSION_NUMBER < 0x10100000 || defined(HAVE_LIBRESSL)
+					n = rsa->n;
+					e = rsa->e;
+					d = rsa->d;
+					p = rsa->p;
+					q = rsa->q;
+					dmp1 = rsa->dmp1;
+					dmq1 = rsa->dmq1;
+					iqmp = rsa->iqmp;
+#else
+					RSA_get0_key(rsa, &n, &e, &d);
+					RSA_get0_factors(rsa, &p, &q);
+					RSA_get0_crt_params(rsa, &dmp1,
+						&dmq1, &iqmp);
+#endif
+					if(!ldns_print_bignum_b64_line(output, "Modulus", n))
 						goto error;
-					}
-					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-						ldns_rdf_deep_free(b64_bignum);
+					if(!ldns_print_bignum_b64_line(output, "PublicExponent", e))
 						goto error;
-					}
-					ldns_rdf_deep_free(b64_bignum);
-					ldns_buffer_printf(output, "\n");
-				} else {
-					ldns_buffer_printf(output, "(Not available)\n");
-				}
-
-				ldns_buffer_printf(output, "Prime1: ");
-				if (rsa->p) {
-					i = (uint16_t)BN_bn2bin(rsa->p, bignum);
-					if (i > LDNS_MAX_KEYLEN) {
+					if(!ldns_print_bignum_b64_line(output, "PrivateExponent", d))
 						goto error;
-					}
-					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-						ldns_rdf_deep_free(b64_bignum);
+					if(!ldns_print_bignum_b64_line(output, "Prime1", p))
 						goto error;
-					}
-					ldns_rdf_deep_free(b64_bignum);
-					ldns_buffer_printf(output, "\n");
-				} else {
-					ldns_buffer_printf(output, "(Not available)\n");
-				}
-
-				ldns_buffer_printf(output, "Prime2: ");
-				if (rsa->q) {
-					i = (uint16_t)BN_bn2bin(rsa->q, bignum);
-					if (i > LDNS_MAX_KEYLEN) {
+					if(!ldns_print_bignum_b64_line(output, "Prime2", q))
 						goto error;
-					}
-					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-						ldns_rdf_deep_free(b64_bignum);
+					if(!ldns_print_bignum_b64_line(output, "Exponent1", dmp1))
 						goto error;
-					}
-					ldns_rdf_deep_free(b64_bignum);
-					ldns_buffer_printf(output, "\n");
-				} else {
-					ldns_buffer_printf(output, "(Not available)\n");
-				}
-
-				ldns_buffer_printf(output, "Exponent1: ");
-				if (rsa->dmp1) {
-					i = (uint16_t)BN_bn2bin(rsa->dmp1, bignum);
-					if (i > LDNS_MAX_KEYLEN) {
+					if(!ldns_print_bignum_b64_line(output, "Exponent2", dmq1))
 						goto error;
-					}
-					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-						ldns_rdf_deep_free(b64_bignum);
+					if(!ldns_print_bignum_b64_line(output, "Coefficient", iqmp))
 						goto error;
-					}
-					ldns_rdf_deep_free(b64_bignum);
-					ldns_buffer_printf(output, "\n");
-				} else {
-					ldns_buffer_printf(output, "(Not available)\n");
-				}
-
-				ldns_buffer_printf(output, "Exponent2: ");
-				if (rsa->dmq1) {
-					i = (uint16_t)BN_bn2bin(rsa->dmq1, bignum);
-					if (i > LDNS_MAX_KEYLEN) {
-						goto error;
-					}
-					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-						ldns_rdf_deep_free(b64_bignum);
-						goto error;
-					}
-					ldns_rdf_deep_free(b64_bignum);
-					ldns_buffer_printf(output, "\n");
-				} else {
-					ldns_buffer_printf(output, "(Not available)\n");
-				}
-
-				ldns_buffer_printf(output, "Coefficient: ");
-				if (rsa->iqmp) {
-					i = (uint16_t)BN_bn2bin(rsa->iqmp, bignum);
-					if (i > LDNS_MAX_KEYLEN) {
-						goto error;
-					}
-					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-						ldns_rdf_deep_free(b64_bignum);
-						goto error;
-					}
-					ldns_rdf_deep_free(b64_bignum);
-					ldns_buffer_printf(output, "\n");
-				} else {
-					ldns_buffer_printf(output, "(Not available)\n");
 				}
 #endif /* splint */
 
@@ -2153,92 +2256,32 @@ ldns_key2buffer_str(ldns_buffer *output, const ldns_key *k)
 
 				/* print to buf, convert to bin, convert to b64,
 				 * print to buf */
-				ldns_buffer_printf(output, "Prime(p): ");
+				if(1) {
+					const BIGNUM *p=NULL, *q=NULL, *g=NULL,
+						*priv_key=NULL, *pub_key=NULL;
+#if OPENSSL_VERSION_NUMBER < 0x10100000 || defined(HAVE_LIBRESSL)
 #ifndef S_SPLINT_S
-				if (dsa->p) {
-					i = (uint16_t)BN_bn2bin(dsa->p, bignum);
-					if (i > LDNS_MAX_KEYLEN) {
-						goto error;
-					}
-					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-						ldns_rdf_deep_free(b64_bignum);
-						goto error;
-					}
-					ldns_rdf_deep_free(b64_bignum);
-					ldns_buffer_printf(output, "\n");
-				} else {
-					printf("(Not available)\n");
-				}
-
-				ldns_buffer_printf(output, "Subprime(q): ");
-				if (dsa->q) {
-					i = (uint16_t)BN_bn2bin(dsa->q, bignum);
-					if (i > LDNS_MAX_KEYLEN) {
-						goto error;
-					}
-					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-						ldns_rdf_deep_free(b64_bignum);
-						goto error;
-					}
-					ldns_rdf_deep_free(b64_bignum);
-					ldns_buffer_printf(output, "\n");
-				} else {
-					printf("(Not available)\n");
-				}
-
-				ldns_buffer_printf(output, "Base(g): ");
-				if (dsa->g) {
-					i = (uint16_t)BN_bn2bin(dsa->g, bignum);
-					if (i > LDNS_MAX_KEYLEN) {
-						goto error;
-					}
-					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-						ldns_rdf_deep_free(b64_bignum);
-						goto error;
-					}
-					ldns_rdf_deep_free(b64_bignum);
-					ldns_buffer_printf(output, "\n");
-				} else {
-					printf("(Not available)\n");
-				}
-
-				ldns_buffer_printf(output, "Private_value(x): ");
-				if (dsa->priv_key) {
-					i = (uint16_t)BN_bn2bin(dsa->priv_key, bignum);
-					if (i > LDNS_MAX_KEYLEN) {
-						goto error;
-					}
-					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-						ldns_rdf_deep_free(b64_bignum);
-						goto error;
-					}
-					ldns_rdf_deep_free(b64_bignum);
-					ldns_buffer_printf(output, "\n");
-				} else {
-					printf("(Not available)\n");
-				}
-
-				ldns_buffer_printf(output, "Public_value(y): ");
-				if (dsa->pub_key) {
-					i = (uint16_t)BN_bn2bin(dsa->pub_key, bignum);
-					if (i > LDNS_MAX_KEYLEN) {
-						goto error;
-					}
-					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-						ldns_rdf_deep_free(b64_bignum);
-						goto error;
-					}
-					ldns_rdf_deep_free(b64_bignum);
-					ldns_buffer_printf(output, "\n");
-				} else {
-					printf("(Not available)\n");
-				}
+					p = dsa->p;
+					q = dsa->q;
+					g = dsa->g;
+					priv_key = dsa->priv_key;
+					pub_key = dsa->pub_key;
 #endif /* splint */
+#else
+					DSA_get0_pqg(dsa, &p, &q, &g);
+					DSA_get0_key(dsa, &pub_key, &priv_key);
+#endif
+					if(!ldns_print_bignum_b64_line(output, "Prime(p)", p))
+						goto error;
+					if(!ldns_print_bignum_b64_line(output, "Subprime(q)", q))
+						goto error;
+					if(!ldns_print_bignum_b64_line(output, "Base(g)", g))
+						goto error;
+					if(!ldns_print_bignum_b64_line(output, "Private_value(x)", priv_key))
+						goto error;
+					if(!ldns_print_bignum_b64_line(output, "Public_value(y)", pub_key))
+						goto error;
+				}
 				break;
 			case LDNS_SIGN_ECC_GOST:
 				/* no format defined, use blob */
@@ -2267,18 +2310,8 @@ ldns_key2buffer_str(ldns_buffer *output, const ldns_key *k)
                                 if(k->_key.key) {
                                         EC_KEY* ec = EVP_PKEY_get1_EC_KEY(k->_key.key);
                                         const BIGNUM* b = EC_KEY_get0_private_key(ec);
-                                        ldns_buffer_printf(output, "PrivateKey: ");
-                                        i = (uint16_t)BN_bn2bin(b, bignum);
-                                        if (i > LDNS_MAX_KEYLEN) {
-                                                goto error;
-                                        }
-                                        b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-                                        if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-						ldns_rdf_deep_free(b64_bignum);
-                                                goto error;
-                                        }
-                                        ldns_rdf_deep_free(b64_bignum);
-				        ldns_buffer_printf(output, "\n");
+					if(!ldns_print_bignum_b64_line(output, "PrivateKey", b))
+						goto error;
                                         /* down reference count in EC_KEY
                                          * its still assigned to the PKEY */
                                         EC_KEY_free(ec);
@@ -2288,6 +2321,28 @@ ldns_key2buffer_str(ldns_buffer *output, const ldns_key *k)
 				goto error;
 #endif /* ECDSA */
                                 break;
+#ifdef USE_ED25519
+			case LDNS_SIGN_ED25519:
+                                ldns_buffer_printf(output, "Private-key-format: v1.2\n");
+				ldns_buffer_printf(output, "Algorithm: %d (", ldns_key_algorithm(k));
+                                status=ldns_algorithm2buffer_str(output, (ldns_algorithm)ldns_key_algorithm(k));
+				ldns_buffer_printf(output, ")\n");
+				if (status) break;
+				status = ldns_ed25519_key2buffer_str(output,
+					k->_key.key);
+				break;
+#endif /* USE_ED25519 */
+#ifdef USE_ED448
+			case LDNS_SIGN_ED448:
+                                ldns_buffer_printf(output, "Private-key-format: v1.2\n");
+				ldns_buffer_printf(output, "Algorithm: %d (", ldns_key_algorithm(k));
+                                status=ldns_algorithm2buffer_str(output, (ldns_algorithm)ldns_key_algorithm(k));
+				ldns_buffer_printf(output, ")\n");
+				if (status) break;
+				status = ldns_ed448_key2buffer_str(output,
+					k->_key.key);
+				break;
+#endif /* USE_ED448 */
 			case LDNS_SIGN_HMACMD5:
 				/* there's not much of a format defined for TSIG */
 				/* It's just a binary blob, Same for all algorithms */
@@ -2300,11 +2355,26 @@ ldns_key2buffer_str(ldns_buffer *output, const ldns_key *k)
 		        ldns_buffer_printf(output, "Algorithm: 158 (HMAC_SHA1)\n");
 				status = ldns_hmac_key2buffer_str(output, k);
 				break;
+			case LDNS_SIGN_HMACSHA224:
+		        ldns_buffer_printf(output, "Private-key-format: v1.2\n");
+		        ldns_buffer_printf(output, "Algorithm: 162 (HMAC_SHA224)\n");
+				status = ldns_hmac_key2buffer_str(output, k);
+				break;
 			case LDNS_SIGN_HMACSHA256:
 		        ldns_buffer_printf(output, "Private-key-format: v1.2\n");
 		        ldns_buffer_printf(output, "Algorithm: 159 (HMAC_SHA256)\n");
 				status = ldns_hmac_key2buffer_str(output, k);
 				break;
+			case LDNS_SIGN_HMACSHA384:
+		        ldns_buffer_printf(output, "Private-key-format: v1.2\n");
+		        ldns_buffer_printf(output, "Algorithm: 164 (HMAC_SHA384)\n");
+				status = ldns_hmac_key2buffer_str(output, k);
+				break;
+			case LDNS_SIGN_HMACSHA512:
+		        ldns_buffer_printf(output, "Private-key-format: v1.2\n");
+		        ldns_buffer_printf(output, "Algorithm: 165 (HMAC_SHA512)\n");
+				status = ldns_hmac_key2buffer_str(output, k);
+				break;
 		}
 #endif /* HAVE_SSL */
 	} else {
