@@ -127,6 +127,9 @@ enum ldns_enum_status {
 	LDNS_STATUS_TYPE_NOT_IN_BITMAP,
 	LDNS_STATUS_INVALID_RDF_TYPE,
 	LDNS_STATUS_RDATA_OVERFLOW,
+	LDNS_STATUS_SYNTAX_SUPERFLUOUS_TEXT_ERR,
+	LDNS_STATUS_NSEC3_DOMAINNAME_OVERFLOW,
+	LDNS_STATUS_DANE_NEED_OPENSSL_GE_1_1_FOR_DANE_TA
 };
 typedef enum ldns_enum_status ldns_status;
 
