@@ -28,6 +28,10 @@
 #include <ldns/rr.h>
 #include <ldns/error.h>
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 /** 
  * DNS Zone
  *
@@ -39,7 +43,7 @@ struct ldns_struct_zone
 {
 	/** the soa defines a zone */
 	ldns_rr 	*_soa;
-	/* basicly a zone is a list of rr's */
+	/* basically a zone is a list of rr's */
 	ldns_rr_list 	*_rrs;
 	/* we could change this to be a b-tree etc etc todo */
 };
@@ -63,7 +67,7 @@ ldns_rr * ldns_zone_soa(const ldns_zone *z);
  * \param[in] z the zone to read from
  * \return the number of rr's in the zone
  */
-uint16_t ldns_zone_rr_count(const ldns_zone *z);
+size_t ldns_zone_rr_count(const ldns_zone *z);
 
 /**
  * Set the zone's soa record
@@ -95,7 +99,7 @@ void ldns_zone_set_rrs(ldns_zone *z, ldns_rr_list *rrlist);
  * \param[in] list the list to add
  * \return a true on succes otherwise falsed
  */
-bool ldns_zone_push_rr_list(ldns_zone *z, ldns_rr_list *list);
+bool ldns_zone_push_rr_list(ldns_zone *z, const ldns_rr_list *list);
 
 /**
  * push an single rr to a zone structure. This function use pointer
@@ -112,7 +116,9 @@ bool ldns_zone_push_rr(ldns_zone *z, ldns_rr *rr);
  * to the zone's data.
  *
  * Due to the current zone implementation (as a list of rr's), this
- * function is extremely slow.
+ * function is extremely slow. Another (probably better) way to do this
+ * is to use an ldns_dnssec_zone structure and the 
+ * ldns_dnssec_mark_and_get_glue() function.
  *
  * \param[in] z the zone to look for glue
  * \return the rr_list with the glue
@@ -129,7 +135,7 @@ ldns_rr_list *ldns_zone_glue_rr_list(const ldns_zone *z);
  *
  * \return ldns_status mesg with an error or LDNS_STATUS_OK
  */
-ldns_status ldns_zone_new_frm_fp(ldns_zone **z, FILE *fp, ldns_rdf *origin, uint16_t ttl, ldns_rr_class c);
+ldns_status ldns_zone_new_frm_fp(ldns_zone **z, FILE *fp, const ldns_rdf *origin, uint32_t ttl, ldns_rr_class c);
 
 /**
  * Create a new zone from a file, keep track of the line numbering
@@ -142,7 +148,7 @@ ldns_status ldns_zone_new_frm_fp(ldns_zone **z, FILE *fp, ldns_rdf *origin, uint
  *
  * \return ldns_status mesg with an error or LDNS_STATUS_OK
  */
-ldns_status ldns_zone_new_frm_fp_l(ldns_zone **z, FILE *fp, ldns_rdf *origin, uint16_t ttl, ldns_rr_class c, int *line_nr);
+ldns_status ldns_zone_new_frm_fp_l(ldns_zone **z, FILE *fp, const ldns_rdf *origin, uint32_t ttl, ldns_rr_class c, int *line_nr);
 
 /**
  * Frees the allocated memory for the zone, and the rr_list structure in it
@@ -163,4 +169,8 @@ void ldns_zone_deep_free(ldns_zone *zone);
  */
 void ldns_zone_sort(ldns_zone *zone);
 
+#ifdef __cplusplus
+}
+#endif
+
 #endif /* LDNS_ZONE_H */
