@@ -202,6 +202,9 @@ ldns_wire2rdf(ldns_rr *rr, const uint8_t *wire, size_t max, size_t *pos)
 			break;
 		case LDNS_RDF_TYPE_CLASS:
 		case LDNS_RDF_TYPE_ALG:
+		case LDNS_RDF_TYPE_CERTIFICATE_USAGE:
+		case LDNS_RDF_TYPE_SELECTOR:
+		case LDNS_RDF_TYPE_MATCHING_TYPE:
 		case LDNS_RDF_TYPE_INT8:
 			cur_rdf_length = LDNS_RDF_SIZE_BYTE;
 			break;
@@ -269,6 +272,7 @@ ldns_wire2rdf(ldns_rr *rr, const uint8_t *wire, size_t max, size_t *pos)
 		case LDNS_RDF_TYPE_ATMA:
 		case LDNS_RDF_TYPE_IPSECKEY:
 		case LDNS_RDF_TYPE_LONG_STR:
+		case LDNS_RDF_TYPE_AMTRELAY:
 		case LDNS_RDF_TYPE_NONE:
 			/*
 			 * Read to end of rr rdata
@@ -389,7 +393,7 @@ ldns_wire2pkt_hdr(ldns_pkt *packet, const uint8_t *wire, size_t max, size_t *pos
 }
 
 ldns_status
-ldns_buffer2pkt_wire(ldns_pkt **packet, ldns_buffer *buffer)
+ldns_buffer2pkt_wire(ldns_pkt **packet, const ldns_buffer *buffer)
 {
 	/* lazy */
 	return ldns_wire2pkt(packet, ldns_buffer_begin(buffer),
@@ -405,10 +409,14 @@ ldns_wire2pkt(ldns_pkt **packet_p, const uint8_t *wire, size_t max)
 	ldns_rr *rr;
 	ldns_pkt *packet = ldns_pkt_new();
 	ldns_status status = LDNS_STATUS_OK;
-	int have_edns = 0;
+	uint8_t have_edns = 0;
 
 	uint8_t data[4];
 
+	if (!packet) {
+		return LDNS_STATUS_MEM_ERR;
+	}
+
 	status = ldns_wire2pkt_hdr(packet, wire, max, &pos);
 	LDNS_STATUS_CHECK_GOTO(status, status_error);
 
@@ -461,6 +469,7 @@ ldns_wire2pkt(ldns_pkt **packet_p, const uint8_t *wire, size_t max)
 			ldns_pkt_set_edns_z(packet, ldns_read_uint16(&data[2]));
 			/* edns might not have rdfs */
 			if (ldns_rr_rdf(rr, 0)) {
+				ldns_rdf_deep_free(ldns_pkt_edns_data(packet));
 				ldns_pkt_set_edns_data(packet, ldns_rdf_clone(ldns_rr_rdf(rr, 0)));
 			}
 			ldns_rr_free(rr);
@@ -477,6 +486,7 @@ ldns_wire2pkt(ldns_pkt **packet_p, const uint8_t *wire, size_t max)
 	if(have_edns)
 		ldns_pkt_set_arcount(packet, ldns_pkt_arcount(packet)
                         - have_edns);
+        packet->_edns_present = have_edns;
 
 	*packet_p = packet;
 	return status;
